/*----------------------------------------------------------------------*/
/*  Name           : algebre.c                                          */
/*  Version        : 1.0                                                */
/*  Creation       : 22/12/95                                           */
/*  Derniere modif.: 14/10/23                                           */
/*  Sujet          : Unites contenant des procedures d'algebre          */
/*  Intitule       : Librairie mathematique supplementaire II           */
/*  Auteur         : Yann Guermeur                                      */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#define true 1
#define false 0
#define sign(a,b) ((b) > 0.0 ? fabs(a) : -fabs(a))
#define NR_END 1
#define FREE_ARG char*
#define small 1e-8

char conf[81];

long dim, dim1, dim2, dim3;

double **mat1, **mat2, **mat3, drand48();


void free_matrix(double **m)

/* Libere l'espace memoire occupe par une matrice */

{

free((double*)m[1]);
free((double*)m);

}

long **matrix_l(int nrow, int ncol)

/* Allocation d'une matrice Cf. Numerical Recipes in C */

{

int ind1, ind2;

long **m;

m = (long **) malloc((size_t)((nrow+1)*sizeof(long*)));
if(!m)
  {
  printf("\nallocation failure 1 in matrix()");
  exit(0);
  }

m[1] = (long *) malloc((size_t)((nrow*(ncol+1))*sizeof(long)));
if(!m[1])
  {
  printf("\nallocation failure 2 in matrix()");
  exit(0);
  }

for(ind1=2;ind1<=nrow;ind1++)
  m[ind1]=m[ind1-1]+ncol;

return m;

}

double **matrix(int nrow, int ncol)

/* Allocation d'une matrice Cf. Numerical Recipes in C */

{

int ind1, ind2;

double **m;

m = (double **) malloc((size_t)((nrow+1)*sizeof(double*)));
if(!m)
  {
  printf("\nallocation failure 1 in matrix()");
  exit(0);
  }

m[1] = (double *) malloc((size_t)((nrow*(ncol+1))*sizeof(double)));
if(!m[1])
  {
  printf("\nallocation failure 2 in matrix()");
  exit(0);
  }

for(ind1=2;ind1<=nrow;ind1++)
  m[ind1]=m[ind1-1]+ncol;

return m;

}

double prod_scal(double *vect1 , double *vect2 , long dim)

/* Produit scalaire de deux vecteurs (de meme dimension) */

{

long ind1;

double produit = 0.0;

for(ind1=1;ind1<=dim;ind1++)
  produit += vect1[ind1] * vect2[ind1];

return produit;

}

double gaussian(double *vect1 , double *vect2 , double dim)

/* Calcul d'une fonction noyau gaussienne */

{

long ind1;

double result=0.0, terme_partiel=0.0;

for(ind1=1; ind1<=dim; ind1++)
  {
  terme_partiel = (vect1[ind1] - vect2[ind1]);
  result -= (terme_partiel * terme_partiel);
  }

result /= (8.0 * (double)dim);

result = exp(result);

return result;

}

double polynomial(double *vect1 , double *vect2 , long dim)

/* Calcul d'une fonction noyau polynomiale */

{

long ind1;

double result=0.0;

for(ind1=1;ind1<=dim;ind1++)
  result += vect1[ind1] * vect2[ind1];
result += 1.0;
result = pow(result, 5.0);

return result;

}

double h_d (double d, double t)

/* Function of Definition 1 in (El Dakdouki et al., 2023) */

{

double remainder, result;

remainder = (t / d - floor ( t / d)) * d;

/* printf("\nt: %lf, d: %lf, remainder: %lf\n", t, d, remainder); */

result = 4 * pow(remainder / d, 2.0) - (4 * remainder / d) + 1.0; 

/* printf("\nvalue taken by function h_d: %lf", result); */

return result;

}

double ker(int nat, double *vect1, double *vect2, long dim)

/* Calcul d'une fonction noyau */

{

double resultat;

/*
printf("\nEntering function ker with nature_kernel = %d\n", nat);
exit(0);
*/

switch(nat)
  {
  case 1: resultat = prod_scal(vect1, vect2, dim); break;
  case 2: resultat = gaussian(vect1, vect2, dim); break;
  case 3: resultat = polynomial(vect1, vect2, dim); break;
  default: printf("\n\nNoyau: %d inconnu...\n\n", nat); exit(0);
  }

return resultat;

}

void display_mat(double **mat, long dim1, long dim2, int par1, int par2)

{

long ind1, ind2;

printf("\n");

for(ind1=1;ind1<=dim1;ind1++)
  {
  for(ind2=1;ind2<=dim2;ind2++)
    printf("%*.*f ",par1,par2,mat[ind1][ind2]);
  printf("\n");
  }

printf("\n");

}

double compute_distance(double *vect1, double *vect2, long dim)

{

long ind1;
double distance = 0.0;

for(ind1=1; ind1<=dim; ind1++)
  distance += pow(vect1[ind1]-vect2[ind1], 2.0);

distance = sqrt(distance);
/* printf("\ndistance = %lf", distance); */
return distance;

}
