/* Copyright 2000-2010 Yann Guermeur                                        */

/* This program is free software; you can redistribute it and/or modify     */
/* it under the terms of the GNU General Public License as published by     */
/* the Free Software Foundation; either version 2 of the License, or        */
/* (at your option) any later version.                                      */

/* This program is distributed in the hope that it will be useful,          */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/* GNU General Public License for more details.                             */

/* You should have received a copy of the GNU General Public License        */
/* along with this program; if not, write to the Free Software              */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/

/*--------------------------------------------------------------------------*/
/*  Name           : train_SVM.c                                            */
/*  Version        : 1.1                                                    */
/*  Creation       : 04/27/00                                               */
/*  Last update    : 12/12/23                                               */
/*  Subject        : Training algorithm of the M-SVM of Weston and Watkins  */
/*  Algo.          : Frank-Wolfe algorithm + decomposition method           */
/*  Author         : Yann Guermeur Yann.Guermeur@loria.fr                   */
/*--------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h> 
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "algebre.h"

#define minimum(a,b) ((a)<=(b)?(a):(b))
#define maximum(a,b) ((a)>=(b)?(a):(b))

#define true 1
#define false 0
#define taille 81
#define step 100
#define very_small 1e-4
#define negliglble 1e-8
#define nb_clust 5

FILE *fs, *fc;

unsigned short xi[3];

int status;

long jump=false, nature_kernel=0, nb_SV=0, row,
i, j, k, l, m, dim_input, nb_data, iter, nb_iter=10000000, *y,
y_i, y_j, Q=0, chunk_size=0, *table_chunk, *in_chunk, ind_pattern,
nrand48(), random_num, rank, active, *out_of_chunk, M;

char fichier_data[taille], fichier_means[taille+1]="Means/",
fichier_alpha[taille], letter, commande[taille+20],
fichier_lp_sol[taille]="lp.res",
*ligne, *frag, fichier_fichcom[taille], fichier_alpha_0[taille],
*tampon[6] = {"00000","0000","000","00","0",""},
fichier_lp[taille]="lp.lp", choice='y';

double **X, **prototypes, **means, **alpha, C=0.0, **gradient, **delta, value,
obj_lp_lu=0.0, obj_lp_comp=0.0, **lp_sol, theta_opt=0.0, **translations,
partiel, **K, **H_delta, gradient_alpha, *mean, *variance, *st_dev,
gradient_Gamma, *rhs, gradient_diff, lambda = 0.1, *diff_vect,
*norms_translations;

/* Functions included in this program */

long indice_tampon(long parametre);
void caract_db();
void alloc_memory();
void read_data();
void standardize_data();
void read_alpha();
void check_feasible_sol();
void k_means();
void compute_translations();
double consolidation();
void compute_table_chunk();
void compute_K();
void compute_gradient();
void write_lp_file();
void read_lp_sol();
void check_opt_sol();
void compute_delta();
void compute_H_delta();
void compute_theta_opt();
void compute_new_alpha();
void write_vector(double **vector, char *fichier);
void Pause(char *message);

int main(int argc, char *argv[])

{

status = system("clear");
srand48(getpid());
strcpy(fichier_fichcom, argv[1]);

caract_db();
read_data();

printf("\nDo you want data to be standardized [y/n]? ");
choice = getc(stdin);
if((choice == 'y') || (choice == 'Y'))
  standardize_data();

read_alpha();
check_feasible_sol();

if(nature_kernel == 4)
  {
  printf("\n\n*** consolidation kernel ***\n");
  k_means();
  compute_translations();
  }

for(iter=1; iter<=nb_iter; iter++)
  {

  if((iter%step) == 0)
    printf("\n\n*** Iteration: %6ld\n", iter);
  jump = false;
  compute_table_chunk();
  compute_K();
  compute_gradient();
  if(jump == false)
    {
    write_lp_file();
    status = system("./lp_solve < lp.lp > lp.res 2> lp.err");
    read_lp_sol();
    }
  if(jump == false)
    check_opt_sol();
  if(jump == false)
    {
    compute_delta();
    compute_theta_opt();
    compute_new_alpha();
    }
  if((iter%step) == 0)
    write_vector(alpha, fichier_alpha);
  } 

}

long indice_tampon(long parametre)

{

long valeur_long;

valeur_long = (long) floor(log10((double) parametre));
return valeur_long;

}

void Pause(char *message)

{

char c;

static int n = 0;

if(message!=NULL)
  printf("%s\n",message);

printf("\a\n=> New Line to proceed...\n");

n = read(0, &c, 1);

while(c != '\n')
  n = read(0, &c, 1);

return;

}

void caract_db()

{

if((fs=fopen(fichier_fichcom, "r"))==NULL)
  {
  printf("\nFile of parameters %s: cannot be open...\n", fichier_fichcom);
  exit(0);
  }

status = fscanf(fs, "%ld", &nature_kernel);
status = fscanf(fs, "%lf", &C);
status = fscanf(fs, "%ld", &chunk_size);

status = fscanf(fs, "%s", fichier_data);
printf("\nThe data file is: %s\n", fichier_data);
status = fscanf(fs, "%s", fichier_alpha_0);
status = fscanf(fs, "%s", fichier_alpha);

fclose(fs);

sprintf(commande, "cp %s Save_alpha/.", fichier_alpha);

}

void alloc_memory()

{

X = matrix(nb_data, dim_input);
mean = (double *) calloc(dim_input+1, sizeof(double));
variance = (double *) calloc(dim_input+1, sizeof(double));
st_dev = (double *) calloc(dim_input+1, sizeof(double));
y = (long *) calloc(nb_data+1, sizeof(long));

if(nature_kernel == 4)
  {
  means = matrix(Q, nb_clust * dim_input);
  prototypes = matrix(nb_clust, dim_input);
  M = Q * nb_clust * (nb_clust - 1) / 2;
  translations = matrix(M, dim_input);
  norms_translations = (double *) calloc(M+1, sizeof(double));
  diff_vect = (double *) calloc(dim_input+1, sizeof(double));
  }

alpha = matrix(nb_data, Q);
in_chunk = (long *) calloc(nb_data+1, sizeof(long));
gradient = matrix(chunk_size, Q);
lp_sol = matrix(chunk_size, Q);
delta = matrix(chunk_size, Q);
table_chunk = (long *) calloc(chunk_size+1, sizeof(long));
out_of_chunk = (long *) calloc(nb_data+1, sizeof(long));
K = matrix(chunk_size, nb_data);
H_delta = matrix(chunk_size, Q);
rhs = (double *) calloc(Q+1, sizeof(double));
ligne = (char *) calloc(taille+1, sizeof(char));
frag = (char *) calloc(taille+1, sizeof(char));

}

void read_data()

{

long min_y, max_y;
double value_y;

if((fs=fopen(fichier_data, "r"))==NULL)
  {
  printf("\nData file %s: cannot be open...\n", fichier_data);
  exit(0);
  }

status = fscanf(fs, "%ld", &nb_data);
status = fscanf(fs, "%ld", &dim_input);
status = fscanf(fs, "%ld", &Q);

min_y = nb_data;
max_y = 0;

alloc_memory();

for(i=1; i<=nb_data; i++)
  {
  for(j=1; j<=dim_input; j++)
    status = fscanf(fs, "%lf", &X[i][j]);
  status = fscanf(fs, "%lf", &value_y);
  y[i] = (long) value_y;
  if(y[i] < min_y)
     min_y = y[i];
  if(y[i] > max_y)
    max_y = y[i];
  }

fclose(fs);

/*
printf("\nExtreme indices of categories : %2ld -> %2ld\n", min_y, max_y);
*/

if(((min_y != 0) && (min_y != 1)) || (max_y - min_y != Q-1))
  {
  printf("\nWrong numbering of the categories\n");
  exit(0);
  }

if(min_y == 0)
  {
  for(i=1; i<=nb_data; i++)
    y[i]++;
  }

}

void standardize_data()

/* Standardize the data of the training set per predictor */

{

for(j=1; j<=dim_input; j++)
  {
  mean[j] = 0.0;
  variance[j] = 0.0;
  st_dev[j] = 0.0;
  }

for(i=1; i<=nb_data; i++)
  for(j=1; j<=dim_input; j++)
    mean[j] += X[i][j];

for(j=1; j<=dim_input; j++)
  {
  mean[j] /= (double) nb_data;
/*  printf("\nmean[%2d] = %e", j, mean[j]); */
  }
/* Pause(""); */

for(i=1; i<=nb_data; i++)
  for(j=1; j<=dim_input; j++)
     variance[j] += (X[i][j] - mean[j]) * (X[i][j] - mean[j]);

for(j=1; j<=dim_input; j++)
  {
  variance[j] /= (double) nb_data;
  st_dev[j] = sqrt(variance[j]);
/*  printf("\nst_dev[%2d] = %e", j, st_dev[j]); */
  }
/* Pause(""); */

for(i=1; i<=nb_data; i++)
  for(j=1; j<=dim_input; j++)
    {
    X[i][j] -= mean[j];
    if(st_dev[j] > negliglble)
      X[i][j] /= st_dev[j];
    }

}

void read_alpha()

{

if((fs=fopen(fichier_alpha_0, "r"))==NULL)
  {
  printf("\nFile of initial dual variable values %s: cannot be open...\n", 
  fichier_alpha_0);
  exit(0);
  }

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    status = fscanf(fs, "%lf", &alpha[i][k]);

fclose(fs);

}

void check_feasible_sol()

{

double *constraints, norm;

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    if((alpha[i][k] < 0.0) || (alpha[i][k] > C))
       {
       printf("\nNo feasible solution: alpha[%ld][%ld] = %lf\n\n",
              i, k, alpha[i][k]);
       exit(0);
       }

constraints = calloc(Q+1, sizeof(double));

for(k=1; k<=Q; k++)
  constraints[k] = 0.0;

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
     {
     if(y[i] == k)
       for(l=1; l<=Q; l++)
         constraints[k] -= alpha[i][l];
     else
       constraints[k] += alpha[i][k];
     }

printf("\nSatisfaction of the equality constraints:\n\n");

for(k=1; k<=Q; k++)
  printf("%11.8f\n", constraints[k]);

norm = 0.0;

for(k=1; k<=Q; k++)
  norm += constraints[k] * constraints[k];

norm = sqrt(norm);

if(norm >= very_small)
  {
  printf("\nLarge deviation of the equality constraints...\n");
  exit(0);
  }

}

void k_means()

{

long *cluster_of_pattern, *cardinal_clust, local_minimum = false;
double dist, dist_min = 0.0, inertia = 0.0, previous_inertia = 0.0, partial;

cluster_of_pattern = (long *) calloc(nb_data+1, sizeof(long));
cardinal_clust = (long *) calloc(nb_clust+1, sizeof(long));

/* printf("\nEntering the k-means function...\n"); */

for(k=1; k<=Q; k++)
  for(j=1; j<=nb_clust * dim_input; j++)
    means[k][j] = 2.0 * drand48() - 1.0;

/*
display_mat(means,Q,nb_clust*dim_input,9,6);
Pause("");
*/

for(k=1; k<=Q; k++)
  {
  for(j=1; j<=nb_clust; j++)
    cardinal_clust[j] = 0;

  for(i=1; i<=nb_clust; i++)
    for(j=1; j<=dim_input; j++)
      prototypes[i][j] = means[k][(i-1)*dim_input+j];

/*
  display_mat(prototypes,nb_clust,dim_input,9,6);
  Pause("");
*/

  for(i=1; i<=nb_data; i++)
    if(y[i] == k)
      {
      cluster_of_pattern[i] = 1;
      dist_min = compute_distance(X[i], prototypes[1], dim_input);
      for(j=2; j<=nb_clust; j++)
        {
        dist = compute_distance(X[i], prototypes[j], dim_input);
        if(dist < dist_min)
          {
          dist_min = dist;
          cluster_of_pattern[i] = j;
          }
	}
      cardinal_clust[cluster_of_pattern[i]]++;
      }

/*
  for(j=1; j<=nb_clust; j++)
    printf("\nFor C=%ld, the cardinal of cluster %ld is %5ld", 
           k, j, cardinal_clust[j]);
*/
  inertia = 0.0;
  for(i=1; i<=nb_data; i++)
    if(y[i] == k)
    {
    partial = compute_distance(X[i],prototypes[cluster_of_pattern[i]],dim_input);
    inertia += pow(partial, 2.0);
    }

  /* printf("\nC=%ld, value of the total inertia: %lf", k, inertia); */
  previous_inertia = inertia;
  local_minimum = false;

  while(local_minimum == false)
    {
    for(i=1; i<=nb_clust; i++)
      if(cardinal_clust[i] != 0)
        for(j=1; j<=dim_input; j++)
          prototypes[i][j] =  0.0;

    for(i=1; i<=nb_data; i++)
      if(y[i] == k)
        for(j=1; j<=dim_input; j++)
          prototypes[cluster_of_pattern[i]][j] += X[i][j];

    for(i=1; i<=nb_clust; i++)
      if(cardinal_clust[i] != 0)
        for(j=1; j<=dim_input; j++)
          prototypes[i][j] /= cardinal_clust[i];

    for(i=1; i<=nb_clust; i++)
      cardinal_clust[i] = 0;

    for(i=1; i<=nb_data; i++)
      if(y[i] == k) 
        {
        cluster_of_pattern[i] = 1;
        dist_min = compute_distance(X[i], prototypes[1], dim_input);
        for(j=2; j<=nb_clust; j++)
          {
          dist = compute_distance(X[i], prototypes[j], dim_input);
          if(dist < dist_min)
            {
            dist_min = dist;
            cluster_of_pattern[i] = j;
            }
          }
        cardinal_clust[cluster_of_pattern[i]]++;
        }

    inertia = 0.0;
    for(i=1; i<=nb_data; i++)
      if(y[i] == k)
        {
        partial=compute_distance(X[i],prototypes[cluster_of_pattern[i]],dim_input);
        inertia += pow(partial, 2.0);
        }

      if(inertia == previous_inertia)
        {
        local_minimum = true;
/*
	if(k < 10)
          printf("\nC=0%1ld, value of the total inertia: %e", k, inertia);
        else
          printf("\nC=%ld, value of the total inertia: %e", k, inertia);
*/
	}
      else
        previous_inertia = inertia;

    }

  for(i=1; i<=nb_clust; i++)
    for(j=1; j<=dim_input; j++)
      means[k][(i-1)*dim_input+j] = prototypes[i][j];
  }

/*
printf("\n\n");
display_mat(means,Q,nb_clust*dim_input,9,6); 
*/

frag = strtok(fichier_data, "/");
frag = strtok(NULL, "/");
strcat(fichier_means, frag);

if((fc=fopen(fichier_means, "w"))==NULL)
  {
  printf("\nFile of means %s: cannot be open...\n", fichier_means);
  exit(0);
  }

for(k=1; k<=Q; k++)
  for(j=1; j<=nb_clust * dim_input; j++)
    fprintf(fc, "%e %c", means[k][j], (j==nb_clust*dim_input) ? '\n': ' ');

fclose(fc);

/* Pause(""); */

}

void compute_translations()

{

long ind_prot_1, ind_prot_2, ind_translation=0;

for(k=1;k<=Q;k++)
  for(ind_prot_1=1; ind_prot_1<nb_clust; ind_prot_1++)
    for(ind_prot_2=ind_prot_1+1; ind_prot_2<=nb_clust; ind_prot_2++)
      { 
      ind_translation++;
      for(j=1; j<=dim_input; j++)   
        translations[ind_translation][j] = 
means[k][(ind_prot_1-1)*dim_input+j] - means[k][(ind_prot_2-1)*dim_input+j];

      norms_translations[ind_translation] = 0.0;
      for(j=1; j<=dim_input; j++)
        norms_translations[ind_translation] 
		+= pow(translations[ind_translation][j], 2.0);
      norms_translations[ind_translation] 
	      = sqrt(norms_translations[ind_translation]);
      }

}

double consolidation(double *vect1 , double *vect2 , long dim)

{

long ind1;

double result=0.0, partial_term=0.0, sum_coefficients=0.0;

for(ind1=1; ind1<=dim; ind1++)
  diff_vect[ind1] = vect1[ind1] - vect2[ind1];

/*
printf("\ndiff vects:");
for(ind1=1; ind1<=dim; ind1++)
  printf("\n %lf", diff_vect[ind1]);
*/

for(ind1=1; ind1<=dim; ind1++)
  result -= (diff_vect[ind1] * diff_vect[ind1]);

result /= (8.0 * (double)dim);

result = exp(result);

for(ind1=1; ind1<=M; ind1++)
  {
  partial_term = prod_scal(translations[ind1], diff_vect, dim);
  partial_term /= norms_translations[ind1];
  partial_term = fabs(partial_term);
  sum_coefficients += h_d (norms_translations[ind1], partial_term);
  }

sum_coefficients /= (double)M;
result = (1 - lambda + lambda * sum_coefficients) * result;

return result;

}

void write_lp_file()

{

if((fc=fopen(fichier_lp, "w"))==NULL)
  {
  printf("\nFile lp %s: cannot be open...\n", fichier_lp);
  exit(0);
  }

fprintf(fc, "min:");

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];
  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      if(gradient[i][k] > 0.0)
        fprintf(fc, " +%lf A0%s%ld", gradient[i][k],
        tampon[indice_tampon(Q*(i-1)+k)], (Q*(i-1)+k));
      if(gradient[i][k] < 0.0) 
        fprintf(fc, " %lf A0%s%ld", gradient[i][k],
        tampon[indice_tampon(Q*(i-1)+k)], (Q*(i-1)+k));
      }
  }

fprintf(fc, ";\n");

for(k=1; k<Q; k++)
  {
  rhs[k] = 0.0;
  for(i=1; i<=nb_data; i++)
    if(in_chunk[i] == 0)
      {
      y_i = y[i];
      for(l=1; l<=Q; l++)
        {
        if((y_i == k) && (l != k))
          rhs[k] += alpha[i][l];
        if((y_i != k) && (l == k))
          rhs[k] -= alpha[i][l];
	}
      }
  }

for(k=1; k<Q; k++)
  {
  fprintf(fc, "CONST%1ld: ", k);

  for(i=1; i<=chunk_size; i++)
    {
    ind_pattern = table_chunk[i];
    y_i = y[ind_pattern];

    if(y_i == k)
      {
      for(l=1; l<=Q; l++)
        {
        if(l != k)
          {
          fprintf(fc, "-A0%s%ld ",
          tampon[indice_tampon(Q*(i-1)+l)],(Q*(i-1)+l));
          }
        }
      }
    else
      fprintf(fc, "+A0%s%ld ",
      tampon[indice_tampon(Q*(i-1)+k)], (Q*(i-1)+k));
    }

  fprintf(fc, "= %lf;\n", rhs[k]);
  }

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];
  for(k=1; k<=Q; k++)
    if(k != y_i)
      fprintf(fc, "A0%s%ld < %lf;\n",
      tampon[indice_tampon(Q*(i-1)+k)], (Q*(i-1)+k), C);
  }

fclose(fc);

}

void compute_table_chunk()

{

for(i=1; i<=nb_data; i++)
  {
  out_of_chunk[i] = i;
  in_chunk[i] = 0;
  }

nb_SV = 0;

for(i=1; i<=nb_data; i++)
  {
  active = 0;
  k = 1;
  while((active == 0) && (k <= Q))
    {
    if((alpha[i][k] > 0.0) && (alpha[i][k] < C))
      {
      active = 1;
      nb_SV++;
      }
    k++;
    }
  }

if((iter%step) == 0)
  printf("\nNumber of margin Sup. Vec.: %ld", nb_SV);

for(i=1; i<=chunk_size; i++)
  {
  if(i==1)
    {
    row = iter % nb_data;
    if(row == 0)
      row = nb_data;
    }
  else
    {
    random_num = nrand48(xi);
    row = (random_num % (nb_data-i+1))+1;
    }
  table_chunk[i] = out_of_chunk[row];
  for(j=row; j<=nb_data-i; j++)
    out_of_chunk[j] = out_of_chunk[j+1];
  in_chunk[table_chunk[i]] = 1;
  }

}

void compute_K()

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  for(j=1; j<=nb_data; j++)
    if(nature_kernel == 4)
      K[i][j] = consolidation(X[ind_pattern], X[j], dim_input);
    else
      K[i][j] = ker(nature_kernel, X[ind_pattern], X[j], dim_input);
  }

}

void compute_gradient()

{

double norm= 0.0;

for(i=1; i<=chunk_size; i++)
  {
  y_i = y[table_chunk[i]];

  for(k=1; k<=Q; k++)
    {
    if(k != y_i)
      {
      gradient[i][k] = -1.0;
      for(j=1; j<=nb_data; j++)
        {
        y_j = y[j];
	partiel = 0.0;

        if(y_j == y_i)
          for(l=1; l<=Q; l++)
            partiel += alpha[j][l];

        if(y_j == k)
          for(l=1; l<=Q; l++)
            partiel -= alpha[j][l];

        partiel += alpha[j][k] - alpha[j][y_i];
	gradient[i][k] += partiel * K[i][j];
        }
      }
    }
  }

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    norm += gradient[i][k] * gradient[i][k];

norm = sqrt(norm);

if(norm < negliglble)
  {  
  if(chunk_size < nb_data)
    jump = true;
  else
    {
    printf("\n\nAn optimal solution has been found !\n\n");
    write_vector(alpha, fichier_alpha);
    exit(0);
    }
  }

}

void compute_H_delta()

{

for(i=1; i<=chunk_size; i++)
  {
  y_i = y[table_chunk[i]];

  for(k=1; k<=Q; k++)
    {
    H_delta[i][k] = 0.0; 
    if(k != y_i)
      {
      for(j=1; j<=chunk_size; j++)
        {
        y_j = y[table_chunk[j]];
        partiel = 0.0;

        if(y_j == y_i)
          for(l=1; l<=Q; l++)
            partiel += delta[j][l];

	if(y_j == k)
	  for(l=1; l<=Q; l++)
            partiel -= delta[j][l];

        partiel += delta[j][k] - delta[j][y_i];
        H_delta[i][k] += partiel * K[i][table_chunk[j]];
	}
      }
    }
  }

}

void read_lp_sol()

{

if((fs=fopen(fichier_lp_sol, "r"))==NULL)
  {
  printf("\nThe file %s of the LP solution cannot be open...\n",
  fichier_lp_sol);
  exit(0);
  }

if(fgets(ligne, taille, fs)==NULL)
  exit(0);
else
  if(strlen(ligne) > 1)
    {
/*
    printf("%s",ligne);
    Pause("");
*/
    jump=true;
    }

if(jump == false)
  {
  if(fgets(ligne, taille, fs)==NULL)
    exit(0);

  if(strncmp("Value of objective function:", ligne, 28) != 0)
    {
/*
    printf("%s",ligne);
    Pause("");
*/
    jump=true;
    }
  else
    {
/*    printf("\nFeasible solution\n"); */
    frag = strtok(ligne, ":");
    frag = strtok(NULL, ":");
    obj_lp_lu = atof(frag);
/*    printf("Value of objective function: %lf\n", obj_lp_lu); */

    if(fgets(ligne, taille, fs)==NULL)
      exit(0);
    if(fgets(ligne, taille, fs)==NULL)
      exit(0);

    for(i=1; i<=chunk_size; i++)
      {
      ind_pattern = table_chunk[i];
      y_i = y[ind_pattern];
      lp_sol[i][y_i]=0.0;
      }

    for(i=1; i<=chunk_size; i++)
      for(k=1; k<Q; k++)
        {
        status = fscanf(fs, "%c", &letter);
        status = fscanf(fs, "%ld", &rank);
        status = fscanf(fs, "%lf", &value);
        j = (rank-1)/Q+1;
        l = (rank-1)%Q+1;
        lp_sol[j][l] = value;
        status = fscanf(fs, "%c", &letter);
        }
    }
  }

fclose(fs);

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    if((lp_sol[i][k] < 0.0) || (lp_sol[i][k] > C))
      jump = true;

}

void check_opt_sol()

{

gradient_alpha = 0.0; 
gradient_Gamma = 0.0;
gradient_diff = 0.0;

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];

  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      gradient_alpha += gradient[i][k] * alpha[ind_pattern][k];
      gradient_Gamma += gradient[i][k] * lp_sol[i][k];
      }
  }

gradient_diff = gradient_alpha - gradient_Gamma;

if(gradient_diff <= 0.0)
  jump = true;

}

void compute_new_alpha()

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  for(k=1; k<=Q; k++)
    {
    alpha[ind_pattern][k] *= (1.0-theta_opt);
    alpha[ind_pattern][k] += theta_opt * lp_sol[i][k];
    }
  }

}

void compute_delta()

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  for(k=1; k<=Q; k++)
    delta[i][k] = alpha[ind_pattern][k] - lp_sol[i][k];
  }

}

void compute_theta_opt()

{

double denominateur = 0.0;

compute_H_delta();

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    denominateur += delta[i][k] * H_delta[i][k];

if(denominateur <=0)
  {
  printf("\nWrong estimate of theta: %e\n", denominateur);
  exit(0);
  }

theta_opt = gradient_diff / denominateur;

if(theta_opt > 1.0)
  theta_opt = 1.0;

/*
if((iter%step) == 0)
  printf("\n    Optimal value of theta: %10.8f\n", theta_opt);
*/

}

void write_vector(double **vector, char *fichier)

{

if((fc=fopen(fichier, "w"))==NULL)
  {
  printf("\nThe file %s cannot be open...\n", fichier);
  exit(0);
  }

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    fprintf(fc, "%e %c", vector[i][k], (k==Q) ? '\n': ' ');

fclose(fc);

status = system(commande);

}
